/* digital clock for X  by Li Wei Jih */

#include <stdio.h>
#include <time.h>
#include <string.h>
#include <unistd.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "position.c"
#include "dclock.xbm"
#include "digit0.xbm"
#include "digit1.xbm"
#include "digit2.xbm"
#include "digit3.xbm"
#include "digit4.xbm"
#include "digit5.xbm"
#include "digit6.xbm"
#include "digit7.xbm"
#include "digit8.xbm"
#include "digit9.xbm"
#include "space.xbm"
#include "colon.xbm"
#include "sec0.xbm"
#include "sec1.xbm"
#include "sec2.xbm"
#include "sec3.xbm"
#include "sec4.xbm"
#include "sec5.xbm"
#include "sec6.xbm"
#include "sec7.xbm"
#include "sec8.xbm"
#include "sec9.xbm"
#include "am.xbm"
#include "pm.xbm"
#include "hr24.xbm"
#include "dash.xbm"
#include "su.xbm"
#include "mo.xbm"
#include "tu.xbm"
#include "we.xbm"
#include "th.xbm"
#include "fr.xbm"
#include "sa.xbm"


char copyright[]= "digital clock for X  by Li Wei Jih";
char appname[]= "clock", iconname[]= "digital clock";
Pixmap clock_pixmap, digit_pixmap[11], colon_pixmap,
       sec_pixmap[10], ampm_pixmap[2], dash_pixmap, day_pixmap[7];
char hr12= 0;
char displayname[255]= "";

static time_t getdigitstr(hr12, timearray)
char hr12;
char *timearray;
{
  time_t cal;
  struct tm *l_time;
  
  cal= time((time_t*) NULL);
  l_time= localtime(&cal);

  if (hr12)
  {					/* 12-hour */
    timearray[6]= l_time->tm_hour>=12;	/* AM/PM */
    l_time->tm_hour= (l_time->tm_hour+11)%12+1;
  }
  else
    timearray[6]= 2;			/* 24-hour */    
  timearray[0]= l_time->tm_hour/10;  timearray[1]= l_time->tm_hour%10;
  timearray[2]= l_time->tm_min/10;   timearray[3]= l_time->tm_min%10;
  timearray[4]= l_time->tm_sec/10;   timearray[5]= l_time->tm_sec%10;
  return (cal);
}

static void showtime(display, window, gc, hr12, timearray)
Display *display;
Window window;
GC gc;
char hr12;
char *timearray;
{
  int i;

  if (timearray[0]==0)
    timearray[0]= 10;

  for (i=0; i<4; i++)
    XCopyPlane(display, digit_pixmap[timearray[i]], window, gc,	/* hour & min */
               0, 0, digit0_width, digit0_height, digit_x[i], digit_y, 1);
  XCopyPlane(display, sec_pixmap[timearray[4]], window, gc,	/* second */
             0, 0, sec0_width, sec0_height, sec_x[0], sec_y, 1);
  XCopyPlane(display, sec_pixmap[timearray[5]], window, gc,
             0, 0, sec0_width, sec0_height, sec_x[1], sec_y, 1);
  XCopyPlane(display, ampm_pixmap[timearray[6]], window, gc,	/* AM/PM */
             0, 0, am_width, am_height, ampm_x, ampm_y, 1);
  XCopyPlane(display, colon_pixmap, window, gc,
             0, 0, colon_width, colon_height, colon_x, colon_y, 1);
}

static void showdate(display, window, gc)
Display *display;
Window window;
GC gc;
{
  time_t cal;
  struct tm *l_time;
  char datearray[4];
  int i;
  
  cal= time((time_t*) NULL);
  l_time= localtime(&cal);
  datearray[0]= (l_time->tm_mon>=9)? 1:10;
  datearray[1]= (l_time->tm_mon+1)%10;
  datearray[2]= l_time->tm_mday/10;  datearray[3]= l_time->tm_mday%10;
  if (datearray[2]==0)
    datearray[2]= 10;
  for (i=0; i<4; i++)
    XCopyPlane(display, digit_pixmap[datearray[i]], window, gc,	/* date */
               0, 0, digit0_width, digit0_height, digit_x[i], digit_y, 1);
  XCopyPlane(display, day_pixmap[l_time->tm_wday], window, gc,	/* day */
             0, 0, su_width, su_height, day_x, day_y, 1);
  XCopyPlane(display, dash_pixmap, window, gc,
             0, 0, dash_width, dash_height, dash_x, dash_y, 1);
  XCopyPlane(display, ampm_pixmap[2], window, gc,
             0, 0, am_width, am_height, ampm_x, ampm_y, 1);
}

void loadpixmap(display, window)
Display *display;
Window window;
{
  char *digit_bits[11], *sec_bits[10], *ampm_bits[3], *day_bits[7];

  int i;
					/* pixmaps to show hour & minute */
  digit_bits[0]= digit0_bits;  digit_bits[1]= digit1_bits;
  digit_bits[2]= digit2_bits;  digit_bits[3]= digit3_bits;
  digit_bits[4]= digit4_bits;  digit_bits[5]= digit5_bits;
  digit_bits[6]= digit6_bits;  digit_bits[7]= digit7_bits;
  digit_bits[8]= digit8_bits;  digit_bits[9]= digit9_bits;
  digit_bits[10]= space_bits;
  for (i=0; i<11; i++)
    digit_pixmap[i]= XCreateBitmapFromData(display, window, digit_bits[i],
                                           digit0_width, digit0_height);

  					/* pixmaps to show second */
  sec_bits[0]= sec0_bits;  sec_bits[1]= sec1_bits;
  sec_bits[2]= sec2_bits;  sec_bits[3]= sec3_bits;
  sec_bits[4]= sec4_bits;  sec_bits[5]= sec5_bits;
  sec_bits[6]= sec6_bits;  sec_bits[7]= sec7_bits;
  sec_bits[8]= sec8_bits;  sec_bits[9]= sec9_bits;
  for (i=0; i<10; i++)
    sec_pixmap[i]= XCreateBitmapFromData(display, window, sec_bits[i],
                                         sec0_width, sec0_height);

  					/* pixmaps to show AM/PM */
  ampm_bits[0]= am_bits;  ampm_bits[1]= pm_bits;  ampm_bits[2]= hr24_bits;
  for (i=0; i<3; i++)
    ampm_pixmap[i]= XCreateBitmapFromData(display, window, ampm_bits[i],
                                          am_width, am_height);

					/* pixmap to show clock */
  clock_pixmap= XCreateBitmapFromData(display, window, dclock_bits,
                                      dclock_width, dclock_height);

					/* pixmap to show colon */
  colon_pixmap= XCreateBitmapFromData(display, window, colon_bits,
                                      colon_width, colon_height);
  dash_pixmap= XCreateBitmapFromData(display, window, dash_bits,
                                      dash_width, dash_height);
  day_bits[0]= su_bits;  day_bits[1]= mo_bits;  day_bits[2]= tu_bits;
  day_bits[3]= we_bits;  day_bits[4]= th_bits;  day_bits[5]= fr_bits;
  day_bits[6]= sa_bits;
  for (i=0; i<7; i++)
    day_pixmap[i]= XCreateBitmapFromData(display, window, day_bits[i],
                                         su_width, su_height);  
}

int args(argc, argv)
int argc;
char **argv;
{
  int i;
  for (i=1; i<argc; i++)
  {
    if (!strcmp(argv[i], "-12"))
    {
      hr12= 1; continue;
    }
    if (!strcmp(argv[i], "-24"))
    {
      hr12= 0; continue;
    }
    if (!strcmp(argv[i], "-display"))
      if (i<argc-1)
      {
        strcpy(displayname, argv[++i]);
        continue;
    }
    puts("Usage: dclock [-12|-24] [-display displayname] [-help]");
    return (1);
  }
  return (0);
}

Window initializewindow(display, window, argc, argv, gc)
Display **display;
Window *window;
int argc;
char **argv;GC *gc;
{
  int screen;
  XSizeHints hint;
  Colormap cmap;
  XColor light;
  unsigned long foreground, background;

  if (args(argc, argv))
    exit(0);

  if (!(*display= XOpenDisplay(displayname)))
  {
    printf("Error: can't open display: %s\n", displayname);
    exit(0);
  }

  hint.width= hint.min_width= hint.max_width= dclock_width;
  hint.height= hint.min_height= hint.max_height= dclock_height;
  hint.flags= PSize | PMinSize | PMaxSize;

  cmap= DefaultColormap(*display, screen= DefaultScreen(*display));
  light.red= 65535; light.green= 65535; light.blue= 0;
  XAllocColor(*display, cmap, &light);  
  background= BlackPixel(*display, screen);
  foreground= light.pixel;		/* bg: black, fg: yellow */

  *window= XCreateSimpleWindow(*display, DefaultRootWindow(*display),
                                 100, 100, hint.width, hint.height,
                                 0, foreground, background);

  loadpixmap(*display, *window);	/* load pixmaps */

  *gc= XCreateGC(*display, *window, 0, 0);
  XSetBackground(*display, *gc, background);
  XSetForeground(*display, *gc, foreground);

  XSetStandardProperties(*display, *window, appname, iconname,
                         clock_pixmap, argv, argc, &hint);

  XSelectInput(*display, *window, ExposureMask | KeyPressMask | KeyReleaseMask |
                                  VisibilityChangeMask | StructureNotifyMask |
                                  ButtonPressMask | Button1MotionMask |
                                  FocusChangeMask);
  XMapRaised(*display, *window);
}


static void processevent(display, window, gc, hr12, timer, timearray)
Display *display;
Window window;
GC gc;
char *hr12;
int *timer;
char *timearray;
{
  while (XEventsQueued(display, QueuedAfterReading))
  {
    XEvent event;
    KeySym key;
    char buffer[8];
    static int org_x, org_y;

    XNextEvent(display, &event);
    switch(event.type)
    {
      case Expose:			/* updates clock display */
        showtime(display, window, gc, *hr12, timearray);
        break;
      case UnmapNotify:
        XMapWindow(display, window);	/* uniconfies clock if iconfied */
        break;
      case ButtonPress:
        if (event.xbutton.button>1)
        {				/* button 2/3 pressed to lower clock */
          XLowerWindow(display, window);
          *timer= 7;			/* lowers clock for 7 sec */
        }
        else
        {				/* button 1 pressed to move clock */
          XRaiseWindow(display, window);
          org_x= event.xbutton.x;	/* gets mouse position */
          org_y= event.xbutton.y;	/* befroe moves clock */
        }
        break;
      case MotionNotify:		/* moves clock */
        XMoveWindow(display, window, 
                    event.xmotion.x_root- org_x, event.xmotion.y_root- org_y);
        break;
      case VisibilityNotify:
        switch (event.xvisibility.state)
        {
          case VisibilityPartiallyObscured:
          case VisibilityFullyObscured:
            if (*timer==0)		/* raises clock if obscured */
              XRaiseWindow(display, window);
            break;
          default:
            getdigitstr(*hr12, timearray);
            showtime(display, window, gc, *hr12, timearray);
        }
        break;
      case KeyPress:
        if (XLookupString(&event.xkey, buffer, sizeof(buffer), &key, NULL))
          switch(buffer[0])
          {
            case ' ':
              showdate(display, window, gc);
              do
                XNextEvent(display, &event);
              while (event.type!=KeyRelease && event.type!=FocusOut);
              break;
            case 't':
            case 'T':
            case '\t':			/* 12/24-hr switch */
              getdigitstr(*hr12= !(*hr12), timearray);
              showtime(display, window, gc, *hr12, timearray);
              break;
            case 'q':
            case 'Q':			/* quits program */
              XFreeGC(display, gc);
              XDestroyWindow(display, window);
              XCloseDisplay(display);
              exit(0);
          }
        break;
    }
  }
}
  

static void runclock(display, window, gc)
Display *display;
Window window;
GC gc;
{
  int timer= 0;  
  char timearray[7];

  getdigitstr(hr12, timearray);
  processevent(display, window, gc, &hr12, &timer, timearray);
  for (;;)
  {
    static time_t l_time= 0;
    
    if (l_time!=time((time_t*) NULL))	/* updates time per second*/
    {
      l_time= getdigitstr(hr12, timearray);
      showtime(display, window, gc, hr12, timearray);
      if (timer>0)
      {
        if (timer--==1)
          XRaiseWindow(display, window);
      }
    }
    XFlush(display);
    usleep(100000L);			/* 1/10 sec */
    processevent(display, window, gc, &hr12, &timer, timearray);    
  }
}

main(argc, argv)
int argc;
char **argv;
{
  Display *display;
  Window window;
  GC gc;
  
  initializewindow(&display, &window, argc, argv, &gc);
  
  runclock(display, window, gc);
}
